
# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.007.
use strict;
use warnings;

use Module::Build 0.28;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.28"
  },
  "configure_requires" => {
    "Module::Build" => "0.28"
  },
  "dist_abstract" => "handles context specific singletons",
  "dist_author" => [
    "Branislav Zahradn\x{ed}k <barney\@cpan.org>"
  ],
  "dist_name" => "Context-Singleton",
  "dist_version" => "1.0.0",
  "license" => "artistic_2",
  "module_name" => "Context::Singleton",
  "recursive_test_files" => 1,
  "requires" => {
    "Class::Load" => 0,
    "Exception::Class" => 0,
    "List::MoreUtils" => 0,
    "List::Util" => 0,
    "Module::Pluggable::Object" => 0,
    "Ref::Util" => 0,
    "Scalar::Util" => 0,
    "Sub::Install" => 0,
    "Sub::Name" => 0,
    "Variable::Magic" => 0,
    "feature" => 0,
    "namespace::clean" => 0,
    "overload" => 0,
    "parent" => 0,
    "strict" => 0,
    "warnings" => 0
  },
  "test_requires" => {
    "Data::Printer" => 0,
    "Exporter" => 0,
    "FindBin" => 0,
    "Hash::Util" => 0,
    "Sub::Override" => 0,
    "Sub::Uplevel" => 0,
    "Test::Deep" => 0,
    "Test::Deep::Cmp" => 0,
    "Test::Exception" => 0,
    "Test::More" => "0.94",
    "lib" => 0
  }
);


my %fallback_build_requires = (
  "Data::Printer" => 0,
  "Exporter" => 0,
  "FindBin" => 0,
  "Hash::Util" => 0,
  "Module::Build" => "0.28",
  "Sub::Override" => 0,
  "Sub::Uplevel" => 0,
  "Test::Deep" => 0,
  "Test::Deep::Cmp" => 0,
  "Test::Exception" => 0,
  "Test::More" => "0.94",
  "lib" => 0
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
